from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.entity import Entity
    from ..models.feature import Feature
    from ..models.feature_view_request_tags import FeatureViewRequestTags


T = TypeVar("T", bound="FeatureViewRequest")


@_attrs_define
class FeatureViewRequest:
    """
    Attributes:
        name (str):
        entities (List['Entity']):
        features (List['Feature']):
        tags (FeatureViewRequestTags):
        description (str):
        ttl (Union[Unset, int]):
    """

    name: str
    entities: List["Entity"]
    features: List["Feature"]
    tags: "FeatureViewRequestTags"
    description: str
    ttl: Union[Unset, int] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        name = self.name

        entities = []
        for entities_item_data in self.entities:
            entities_item = entities_item_data.to_dict()
            entities.append(entities_item)

        features = []
        for features_item_data in self.features:
            features_item = features_item_data.to_dict()
            features.append(features_item)

        tags = self.tags.to_dict()

        description = self.description

        ttl = self.ttl

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "name": name,
                "entities": entities,
                "features": features,
                "tags": tags,
                "description": description,
            }
        )
        if ttl is not UNSET:
            field_dict["ttl"] = ttl

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.entity import Entity
        from ..models.feature import Feature
        from ..models.feature_view_request_tags import FeatureViewRequestTags

        d = src_dict.copy()
        name = d.pop("name")

        entities = []
        _entities = d.pop("entities")
        for entities_item_data in _entities:
            entities_item = Entity.from_dict(entities_item_data)

            entities.append(entities_item)

        features = []
        _features = d.pop("features")
        for features_item_data in _features:
            features_item = Feature.from_dict(features_item_data)

            features.append(features_item)

        tags = FeatureViewRequestTags.from_dict(d.pop("tags"))

        description = d.pop("description")

        ttl = d.pop("ttl", UNSET)

        feature_view_request = cls(
            name=name,
            entities=entities,
            features=features,
            tags=tags,
            description=description,
            ttl=ttl,
        )

        feature_view_request.additional_properties = d
        return feature_view_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains all the data models used in inputs/outputs"""

from .datasource_config import DatasourceConfig
from .datasource_dto import DatasourceDto
from .datasource_dto_added_by import DatasourceDtoAddedBy
from .datasource_dto_auth_type import DatasourceDtoAuthType
from .datasource_dto_data_source_type import DatasourceDtoDataSourceType
from .datasource_dto_status import DatasourceDtoStatus
from .datasource_owner_info import DatasourceOwnerInfo
from .error_response import ErrorResponse
from .key_request import KeyRequest
from .list_request import ListRequest
from .log_metric_m import LogMetricM
from .log_metric_t import LogMetricT
from .proxy_error_response import ProxyErrorResponse

__all__ = (
    "DatasourceConfig",
    "DatasourceDto",
    "DatasourceDtoAddedBy",
    "DatasourceDtoAuthType",
    "DatasourceDtoDataSourceType",
    "DatasourceDtoStatus",
    "DatasourceOwnerInfo",
    "ErrorResponse",
    "KeyRequest",
    "ListRequest",
    "LogMetricM",
    "LogMetricT",
    "ProxyErrorResponse",
)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.create_training_set_version_request_meta import (
        CreateTrainingSetVersionRequestMeta,
    )
    from ..models.monitoring_meta import MonitoringMeta


T = TypeVar("T", bound="CreateTrainingSetVersionRequest")


@_attrs_define
class CreateTrainingSetVersionRequest:
    """
    Attributes:
        project_id (str):
        key_columns (List[str]):
        target_columns (List[str]):
        exclude_columns (List[str]):
        all_columns (List[str]):
        monitoring_meta (MonitoringMeta):
        meta (CreateTrainingSetVersionRequestMeta):
        description (Union[Unset, str]):
    """

    project_id: str
    key_columns: List[str]
    target_columns: List[str]
    exclude_columns: List[str]
    all_columns: List[str]
    monitoring_meta: "MonitoringMeta"
    meta: "CreateTrainingSetVersionRequestMeta"
    description: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        project_id = self.project_id

        key_columns = self.key_columns

        target_columns = self.target_columns

        exclude_columns = self.exclude_columns

        all_columns = self.all_columns

        monitoring_meta = self.monitoring_meta.to_dict()

        meta = self.meta.to_dict()

        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "projectId": project_id,
                "keyColumns": key_columns,
                "targetColumns": target_columns,
                "excludeColumns": exclude_columns,
                "allColumns": all_columns,
                "monitoringMeta": monitoring_meta,
                "meta": meta,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.create_training_set_version_request_meta import (
            CreateTrainingSetVersionRequestMeta,
        )
        from ..models.monitoring_meta import MonitoringMeta

        d = src_dict.copy()
        project_id = d.pop("projectId")

        key_columns = cast(List[str], d.pop("keyColumns"))

        target_columns = cast(List[str], d.pop("targetColumns"))

        exclude_columns = cast(List[str], d.pop("excludeColumns"))

        all_columns = cast(List[str], d.pop("allColumns"))

        monitoring_meta = MonitoringMeta.from_dict(d.pop("monitoringMeta"))

        meta = CreateTrainingSetVersionRequestMeta.from_dict(d.pop("meta"))

        description = d.pop("description", UNSET)

        create_training_set_version_request = cls(
            project_id=project_id,
            key_columns=key_columns,
            target_columns=target_columns,
            exclude_columns=exclude_columns,
            all_columns=all_columns,
            monitoring_meta=monitoring_meta,
            meta=meta,
            description=description,
        )

        create_training_set_version_request.additional_properties = d
        return create_training_set_version_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="MonitoringMeta")


@_attrs_define
class MonitoringMeta:
    """
    Attributes:
        timestamp_columns (List[str]):
        categorical_columns (List[str]):
        ordinal_columns (List[str]):
    """

    timestamp_columns: List[str]
    categorical_columns: List[str]
    ordinal_columns: List[str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        timestamp_columns = self.timestamp_columns

        categorical_columns = self.categorical_columns

        ordinal_columns = self.ordinal_columns

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "timestampColumns": timestamp_columns,
                "categoricalColumns": categorical_columns,
                "ordinalColumns": ordinal_columns,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        timestamp_columns = cast(List[str], d.pop("timestampColumns"))

        categorical_columns = cast(List[str], d.pop("categoricalColumns"))

        ordinal_columns = cast(List[str], d.pop("ordinalColumns"))

        monitoring_meta = cls(
            timestamp_columns=timestamp_columns,
            categorical_columns=categorical_columns,
            ordinal_columns=ordinal_columns,
        )

        monitoring_meta.additional_properties = d
        return monitoring_meta

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ConfigFieldName(str, Enum):
    ACCOUNTID = "accountId"
    AUTHENTICATOR = "authenticator"
    BLOBEXPORTLOCATION = "blobExportLocation"
    CLUSTERID = "clusterId"
    CONFIGPATH = "configPath"
    DATABASE = "database"
    DATASETNAME = "datasetName"
    GCSPROJECTNAME = "gcsProjectName"
    GCSSTAGINGLOCATION = "gcsStagingLocation"
    IAMROLE = "iamRole"
    LOCATION = "location"
    REGION = "region"
    ROLE = "role"
    S3STAGINGLOCATION = "s3StagingLocation"
    SCHEMA = "schema"
    STORAGEINTEGRATIONNAME = "storageIntegrationName"
    WAREHOUSE = "warehouse"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.feature_view_request import FeatureViewRequest


T = TypeVar("T", bound="UpsertFeatureViewsRequest")


@_attrs_define
class UpsertFeatureViewsRequest:
    """
    Attributes:
        feature_views (List['FeatureViewRequest']):
        git_commit_hash (str):
        project_id (str):
    """

    feature_views: List["FeatureViewRequest"]
    git_commit_hash: str
    project_id: str
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        feature_views = []
        for feature_views_item_data in self.feature_views:
            feature_views_item = feature_views_item_data.to_dict()
            feature_views.append(feature_views_item)

        git_commit_hash = self.git_commit_hash

        project_id = self.project_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "featureViews": feature_views,
                "gitCommitHash": git_commit_hash,
                "projectId": project_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.feature_view_request import FeatureViewRequest

        d = src_dict.copy()
        feature_views = []
        _feature_views = d.pop("featureViews")
        for feature_views_item_data in _feature_views:
            feature_views_item = FeatureViewRequest.from_dict(feature_views_item_data)

            feature_views.append(feature_views_item)

        git_commit_hash = d.pop("gitCommitHash")

        project_id = d.pop("projectId")

        upsert_feature_views_request = cls(
            feature_views=feature_views,
            git_commit_hash=git_commit_hash,
            project_id=project_id,
        )

        upsert_feature_views_request.additional_properties = d
        return upsert_feature_views_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class OfflineStoreType(str, Enum):
    BIGQUERY = "BigQuery"
    FILE = "File"
    REDSHIFT = "Redshift"
    SNOWFLAKE = "Snowflake"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.training_set_version_filter_meta import TrainingSetVersionFilterMeta
    from ..models.training_set_version_filter_training_set_meta import (
        TrainingSetVersionFilterTrainingSetMeta,
    )


T = TypeVar("T", bound="TrainingSetVersionFilter")


@_attrs_define
class TrainingSetVersionFilter:
    """
    Attributes:
        project_id (str):
        training_set_meta (TrainingSetVersionFilterTrainingSetMeta):
        meta (TrainingSetVersionFilterMeta):
        training_set_name (Union[Unset, str]):
    """

    project_id: str
    training_set_meta: "TrainingSetVersionFilterTrainingSetMeta"
    meta: "TrainingSetVersionFilterMeta"
    training_set_name: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        project_id = self.project_id

        training_set_meta = self.training_set_meta.to_dict()

        meta = self.meta.to_dict()

        training_set_name = self.training_set_name

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "projectId": project_id,
                "trainingSetMeta": training_set_meta,
                "meta": meta,
            }
        )
        if training_set_name is not UNSET:
            field_dict["trainingSetName"] = training_set_name

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.training_set_version_filter_meta import TrainingSetVersionFilterMeta
        from ..models.training_set_version_filter_training_set_meta import (
            TrainingSetVersionFilterTrainingSetMeta,
        )

        d = src_dict.copy()
        project_id = d.pop("projectId")

        training_set_meta = TrainingSetVersionFilterTrainingSetMeta.from_dict(
            d.pop("trainingSetMeta")
        )

        meta = TrainingSetVersionFilterMeta.from_dict(d.pop("meta"))

        training_set_name = d.pop("trainingSetName", UNSET)

        training_set_version_filter = cls(
            project_id=project_id,
            training_set_meta=training_set_meta,
            meta=meta,
            training_set_name=training_set_name,
        )

        training_set_version_filter.additional_properties = d
        return training_set_version_filter

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, Optional, Union

from http import HTTPStatus

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.log_metric_m import LogMetricM
from ...models.log_metric_t import LogMetricT
from ...types import UNSET, Response


def _get_kwargs(
    *,
    t: LogMetricT,
    b: int,
    m: LogMetricM,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}

    json_t = t.value
    params["t"] = json_t

    params["b"] = b

    json_m = m.value
    params["m"] = json_m

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "head",
        "url": "/objectstore/metric",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Any]:
    if response.status_code == 200:
        return None
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Any]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    t: LogMetricT,
    b: int,
    m: LogMetricM,
) -> Response[Any]:
    """Log metrics about file size read or written

    Args:
        t (LogMetricT):
        b (int):
        m (LogMetricM):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Any]
    """

    kwargs = _get_kwargs(
        t=t,
        b=b,
        m=m,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    t: LogMetricT,
    b: int,
    m: LogMetricM,
) -> Response[Any]:
    """Log metrics about file size read or written

    Args:
        t (LogMetricT):
        b (int):
        m (LogMetricM):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Any]
    """

    kwargs = _get_kwargs(
        t=t,
        b=b,
        m=m,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)



"""Contains all the data models used in inputs/outputs"""

from .auth_config import AuthConfig
from .auth_config_fields import AuthConfigFields
from .auth_config_meta import AuthConfigMeta
from .auth_field_name import AuthFieldName
from .auth_type import AuthType
from .config_field_name import ConfigFieldName
from .create_feature_store_request import CreateFeatureStoreRequest
from .create_feature_store_request_offline_store_config import (
    CreateFeatureStoreRequestOfflineStoreConfig,
)
from .entity import Entity
from .feature import Feature
from .feature_store import FeatureStore
from .feature_store_offline_store_config import FeatureStoreOfflineStoreConfig
from .feature_store_sync_result import FeatureStoreSyncResult
from .feature_tags import FeatureTags
from .feature_view import FeatureView
from .feature_view_request import FeatureViewRequest
from .feature_view_request_tags import FeatureViewRequestTags
from .feature_view_tags import FeatureViewTags
from .field import Field
from .git_provider_name import GitProviderName
from .lock_feature_store_request import LockFeatureStoreRequest
from .metadata import Metadata
from .offline_store_config import OfflineStoreConfig
from .offline_store_config_fields import OfflineStoreConfigFields
from .offline_store_type import OfflineStoreType
from .unlock_feature_store_request import UnlockFeatureStoreRequest
from .upsert_feature_views_request import UpsertFeatureViewsRequest

__all__ = (
    "AuthConfig",
    "AuthConfigFields",
    "AuthConfigMeta",
    "AuthFieldName",
    "AuthType",
    "ConfigFieldName",
    "CreateFeatureStoreRequest",
    "CreateFeatureStoreRequestOfflineStoreConfig",
    "Entity",
    "Feature",
    "FeatureStore",
    "FeatureStoreOfflineStoreConfig",
    "FeatureStoreSyncResult",
    "FeatureTags",
    "FeatureView",
    "FeatureViewRequest",
    "FeatureViewRequestTags",
    "FeatureViewTags",
    "Field",
    "GitProviderName",
    "LockFeatureStoreRequest",
    "Metadata",
    "OfflineStoreConfig",
    "OfflineStoreConfigFields",
    "OfflineStoreType",
    "UnlockFeatureStoreRequest",
    "UpsertFeatureViewsRequest",
)



from typing import Any, Dict, List, Optional, Union

from http import HTTPStatus

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.training_set import TrainingSet
from ...models.training_set_filter import TrainingSetFilter
from ...types import UNSET, Response, Unset


def _get_kwargs(
    *,
    body: TrainingSetFilter,
    offset: Union[Unset, int] = UNSET,
    limit: Union[Unset, int] = UNSET,
    asc: Union[Unset, bool] = UNSET,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    params: Dict[str, Any] = {}

    params["offset"] = offset

    params["limit"] = limit

    params["asc"] = asc

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "post",
        "url": "/find",
        "params": params,
    }

    _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[List["TrainingSet"]]:
    if response.status_code == HTTPStatus.OK:
        response_200 = []
        _response_200 = response.json()
        for response_200_item_data in _response_200:
            response_200_item = TrainingSet.from_dict(response_200_item_data)

            response_200.append(response_200_item)

        return response_200
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[List["TrainingSet"]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    body: TrainingSetFilter,
    offset: Union[Unset, int] = UNSET,
    limit: Union[Unset, int] = UNSET,
    asc: Union[Unset, bool] = UNSET,
) -> Response[List["TrainingSet"]]:
    """List TrainingSets

    Args:
        offset (Union[Unset, int]):
        limit (Union[Unset, int]):
        asc (Union[Unset, bool]):
        body (TrainingSetFilter):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[List['TrainingSet']]
    """

    kwargs = _get_kwargs(
        body=body,
        offset=offset,
        limit=limit,
        asc=asc,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    body: TrainingSetFilter,
    offset: Union[Unset, int] = UNSET,
    limit: Union[Unset, int] = UNSET,
    asc: Union[Unset, bool] = UNSET,
) -> Optional[List["TrainingSet"]]:
    """List TrainingSets

    Args:
        offset (Union[Unset, int]):
        limit (Union[Unset, int]):
        asc (Union[Unset, bool]):
        body (TrainingSetFilter):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        List['TrainingSet']
    """

    return sync_detailed(
        client=client,
        body=body,
        offset=offset,
        limit=limit,
        asc=asc,
    ).parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    body: TrainingSetFilter,
    offset: Union[Unset, int] = UNSET,
    limit: Union[Unset, int] = UNSET,
    asc: Union[Unset, bool] = UNSET,
) -> Response[List["TrainingSet"]]:
    """List TrainingSets

    Args:
        offset (Union[Unset, int]):
        limit (Union[Unset, int]):
        asc (Union[Unset, bool]):
        body (TrainingSetFilter):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[List['TrainingSet']]
    """

    kwargs = _get_kwargs(
        body=body,
        offset=offset,
        limit=limit,
        asc=asc,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    body: TrainingSetFilter,
    offset: Union[Unset, int] = UNSET,
    limit: Union[Unset, int] = UNSET,
    asc: Union[Unset, bool] = UNSET,
) -> Optional[List["TrainingSet"]]:
    """List TrainingSets

    Args:
        offset (Union[Unset, int]):
        limit (Union[Unset, int]):
        asc (Union[Unset, bool]):
        body (TrainingSetFilter):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        List['TrainingSet']
    """

    return (
        await asyncio_detailed(
            client=client,
            body=body,
            offset=offset,
            limit=limit,
            asc=asc,
        )
    ).parsed



"""Containing script to sync feature store from feast to domino"""

import argparse
import sys

from .logging import logger
from .sync import feature_store_sync

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Run feature store sync.")
    parser.add_argument(
        "-i",
        "--id",
        metavar="FEATURE_STORE_ID",
        type=str,
        required=True,
        help="the domino feature store id",
    )
    parser.add_argument(
        "-p",
        "--path",
        metavar="FEAST_REPO_PATH",
        type=str,
        required=False,
        help="the feature store git repo path",
    )
    parser.add_argument(
        "-b",
        "--branch",
        metavar="FEAST_REPO_BRANCH",
        type=str,
        required=False,
        help="the feature store git repo branch",
    )

    parser.add_argument(
        "-r",
        "--retry",
        metavar="MAX_LOCK_RETRY_COUNT",
        type=int,
        required=False,
        help="the maximum feature store lock retry count",
    )
    args = parser.parse_args()

    try:
        feature_store_sync(args.id, args.path, args.branch, args.retry)
    except Exception as e:
        logger.exception(e)
        sys.exit(-1)



"""Datasource module."""

from typing import Any, Dict, List, Optional, cast

import configparser
import json
import os

import attr
import backoff
import httpx
import pandas
import urllib3
from httpx._config import DEFAULT_TIMEOUT_CONFIG
from pyarrow import ArrowException, flight, parquet

import domino_data.configuration_gen
from datasource_api_client.api.datasource import get_datasource_by_name
from datasource_api_client.api.proxy import get_key_url, list_keys, log_metric
from datasource_api_client.models import DatasourceConfig as APIConfig
from datasource_api_client.models import (
    DatasourceDto,
    DatasourceDtoAuthType,
    DatasourceDtoDataSourceType,
    ErrorResponse,
    KeyRequest,
    ListRequest,
    LogMetricM,
    LogMetricT,
    ProxyErrorResponse,
)

from .auth import AuthenticatedClient, AuthMiddlewareFactory, ProxyClient, get_jwt_token
from .configuration_gen import Config, CredElem, DatasourceConfig, find_datasource_klass
from .logging import logger
from .meta import MetaMiddlewareFactory
from .transfer import MAX_WORKERS, BlobTransfer

ACCEPT_HEADERS = {"Accept": "application/json"}
ADLS_HEADERS = {"X-Ms-Blob-Type": "BlockBlob"}

CREDENTIAL_TYPE = "credential"
CONFIGURATION_TYPE = "configuration"

FLIGHT_ERROR_SPLIT = ". gRPC client debug context:"

AWS_CREDENTIALS_DEFAULT_LOCATION = "/var/lib/domino/home/.aws/credentials"
AWS_SHARED_CREDENTIALS_FILE = "AWS_SHARED_CREDENTIALS_FILE"
DOMINO_API_HOST = "DOMINO_API_HOST"
DOMINO_API_PROXY = "DOMINO_API_PROXY"
DOMINO_CLIENT_SOURCE = "DOMINO_CLIENT_SOURCE"
DOMINO_DATASOURCE_PROXY_HOST = "DOMINO_DATASOURCE_PROXY_HOST"
DOMINO_DATASOURCE_PROXY_FLIGHT_HOST = "DOMINO_DATASOURCE_PROXY_FLIGHT_HOST"
DOMINO_RUN_ID = "DOMINO_RUN_ID"
DOMINO_USER_API_KEY = "DOMINO_USER_API_KEY"
DOMINO_USER_HOST = "DOMINO_USER_HOST"
DOMINO_TOKEN_DEFAULT_LOCATION = "/var/lib/domino/home/.api/token"
DOMINO_TOKEN_FILE = "DOMINO_TOKEN_FILE"


def __getattr__(name: str) -> Any:
    if name.endswith("Config"):
        return getattr(domino_data.configuration_gen, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


def __dir__() -> List[str]:
    confs = filter(lambda x: x.endswith("Config"), dir(domino_data.configuration_gen))
    return list(globals().keys()) + list(confs)


class DominoError(Exception):
    """Base exception for known errors."""


class UnauthenticatedError(DominoError):
    """To handle exponential backoff."""


def _unpack_flight_error(error: str) -> str:
    """Unpack a flight error message by removing extra information."""
    try:
        return error.split(FLIGHT_ERROR_SPLIT, maxsplit=1)[0]
    except ValueError:
        return error


@attr.s
class Result:
    """Represents a query result."""

    client: "DataSourceClient" = attr.ib(repr=False)
    reader: flight.FlightStreamReader = attr.ib(repr=False)
    statement: str = attr.ib()

    def to_pandas(self) -> pandas.DataFrame:
        """Load and transform the result into a pandas DataFrame.

        Returns:
            Pandas dataframe loaded with entire resultset
        """
        return self.reader.read_pandas()

    def to_parquet(self, where: Any) -> None:
        """Load and serialize the result to a local parquet file.

        Args:
            where: path of file-like object.
        """
        table = self.reader.read_all()
        # In Pyarrow v13.0, the parquet version was upgraded to v2.6 from v2.4.
        # Set the coerce_timestamps to "us"(microseconds) for backward compatibility.
        parquet.write_table(table, where, coerce_timestamps="us")


@attr.s
class _Object:
    """Represents an object in a object store."""

    datasource: "ObjectStoreDatasource" = attr.ib(repr=False)
    key: str = attr.ib()

    def http(self) -> httpx.Client:
        """Get datasource http client."""
        return self.datasource.http()

    def pool_manager(self) -> urllib3.PoolManager:
        """Get datasource http pool manager."""
        return self.datasource.pool_manager()

    def get(self) -> bytes:
        """Get object content as bytes."""
        url = self.datasource.get_key_url(self.key, False)
        res = self.http().get(url)
        res.raise_for_status()

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            len(res.content),
            False,
        )

        return res.content

    def download_file(self, filename: str) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exists.

        Args:
            filename: path of file to write content to.
        """
        url = self.datasource.get_key_url(self.key, False)
        content_size = 0
        with self.http().stream("GET", url) as stream, open(filename, "wb") as file:
            for data in stream.iter_bytes():
                content_size += len(data)
                file.write(data)

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            content_size,
            False,
        )

    def download(self, filename: str, max_workers: int = MAX_WORKERS) -> None:
        """Download object content to file with multithreaded support.

        The file will be created if it does not exists. File will be overwritten if it exists.

        Args:
            filename: path of file to write content to
            max_workers: max parallelism for high speed download
        """
        url = self.datasource.get_key_url(self.key, False)
        with open(filename, "wb") as file:
            blob = BlobTransfer(url, file, max_workers=max_workers, http=self.pool_manager())

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            blob.content_size,
            False,
        )

    def download_fileobj(self, fileobj: Any) -> None:
        """Download object content to file like object.

        Args:
            fileobj: A file-like object to download into.
                At a minimum, it must implement the write method and must accept bytes.
        """
        url = self.datasource.get_key_url(self.key, False)
        content_size = 0
        with self.http().stream("GET", url) as stream:
            for data in stream.iter_bytes():
                content_size += len(data)
                fileobj.write(data)

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            content_size,
            False,
        )

    def put(self, content: bytes) -> None:
        """Upload content to object.

        Args:
            content: bytes content
        """
        url = self.datasource.get_key_url(self.key, True)
        res = self.http().put(url, content=content)
        res.raise_for_status()

        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            len(content),
            True,
        )

    def upload_file(self, filename: str) -> None:
        """Upload content of file at filename to object.

        Args:
            filename: path of file to upload.
        """
        url = self.datasource.get_key_url(self.key, True)
        with open(filename, "rb") as file:
            res = self.http().put(url, content=file)
        res.raise_for_status()

        content_size = os.path.getsize(filename)
        self.datasource.client._log_metric(  # pylint: disable=protected-access
            self.datasource.datasource_type,
            content_size,
            True,
        )

    def upload_fileobj(self, fileobj: Any) -> None:
        """Upload content of file like object to object.

        Args:
            fileobj: bytes-like object or an iterable producing bytes.
        """
        url = self.datasource.get_key_url(self.key, True)
        res = self.http().put(url, content=fileobj)
        res.raise_for_status()


def load_oauth_credentials() -> Dict[str, str]:
    """Load oauth token from sidecar container or local file.

    Returns:
        .. code-block:: python

            {CredElem.TOKEN.value: "token"}

    Raises:
        DominoError: if the provided location is not a valid file
    """
    token_url = os.getenv(DOMINO_API_PROXY, "")
    if token_url:
        try:
            jwt = get_jwt_token(token_url)
        except httpx.HTTPStatusError:
            logger.opt(exception=True).warning("Failed to get token from sidecar container API")
        else:
            return {CredElem.TOKEN.value: jwt}

    location = os.getenv("DOMINO_TOKEN_FILE", DOMINO_TOKEN_DEFAULT_LOCATION)
    try:
        with open(location, encoding="ascii") as token_file:
            return {CredElem.TOKEN.value: token_file.readline().rstrip()}
    except FileNotFoundError as exc:
        raise DominoError("Token file location is not a valid file.") from exc


def load_aws_credentials(location: str, profile: str = "") -> Dict[str, str]:
    """Load AWS credentials from given location and profile.

    Args:
        location: location of file that contains token.
        profile: profile to load.

    Returns:
        .. code-block:: python

            {
                CredElem.ACCESSKEYID.value: "access_key_id",
                CredElem.SECRETACCESSKEY.value: "secret_access_key",
                CredElem.SESSIONTOKEN.value: "session_token",
            }

    Raises:
        DominoError: if the provided location is not a valid file
    """
    location = location or AWS_CREDENTIALS_DEFAULT_LOCATION
    aws_config = configparser.RawConfigParser()
    aws_config.read(location)
    if not aws_config or not aws_config.sections():
        raise DominoError("AWS credentials file does not exist or does not contain profiles")

    profile = profile or aws_config.sections()[0]
    return {
        CredElem.ACCESSKEYID.value: aws_config.get(profile, "aws_access_key_id"),
        CredElem.SECRETACCESSKEY.value: aws_config.get(profile, "aws_secret_access_key"),
        CredElem.SESSIONTOKEN.value: aws_config.get(profile, "aws_session_token"),
    }


@attr.s
class Datasource:
    """Represents a Domino datasource."""

    # pylint: disable=too-many-instance-attributes

    auth_type: str = attr.ib()
    client: "DataSourceClient" = attr.ib(repr=False)
    config: Dict[str, Any] = attr.ib()
    datasource_type: str = attr.ib()
    identifier: str = attr.ib()
    name: str = attr.ib()
    owner: str = attr.ib()

    _config_override: DatasourceConfig = attr.ib(factory=Config, init=False, repr=False)
    _httpx: Optional[httpx.Client] = attr.ib(None, init=False, repr=False)

    @classmethod
    def from_dto(cls, client: "DataSourceClient", dto: DatasourceDto) -> "Datasource":
        """Build a datasource from a given DTO."""
        return cls(
            auth_type=dto.auth_type.value,
            client=client,
            config=dto.config.to_dict(),
            datasource_type=dto.data_source_type.value,
            identifier=dto.id,
            name=dto.name,
            owner=dto.owner_info.owner_name,
        )

    def http(self) -> httpx.Client:
        """Singleton http client built for the datasource."""
        if self._httpx is not None:
            return self._httpx

        context = httpx.create_ssl_context()

        if self.datasource_type in (
            DatasourceDtoDataSourceType.ADLSCONFIG.value,
            DatasourceDtoDataSourceType.AZUREBLOBSTORAGECONFIG.value,
        ):
            self._httpx = httpx.Client(
                headers=ADLS_HEADERS, verify=context, timeout=DEFAULT_TIMEOUT_CONFIG
            )
        elif self.datasource_type == DatasourceDtoDataSourceType.GENERICS3CONFIG.value:
            self._httpx = httpx.Client(verify=False, timeout=DEFAULT_TIMEOUT_CONFIG)  # nosec
        else:
            self._httpx = httpx.Client(verify=context, timeout=DEFAULT_TIMEOUT_CONFIG)
        return self._httpx

    def pool_manager(self) -> urllib3.PoolManager:
        """Urllib3 pool manager for range downloads."""
        if self.datasource_type in (
            DatasourceDtoDataSourceType.ADLSCONFIG.value,
            DatasourceDtoDataSourceType.AZUREBLOBSTORAGECONFIG.value,
        ):
            return urllib3.PoolManager(headers=ADLS_HEADERS)
        elif self.datasource_type == DatasourceDtoDataSourceType.GENERICS3CONFIG.value:
            return urllib3.PoolManager(assert_hostname=False)
        else:
            return urllib3.PoolManager()

    def _get_credential_override(self) -> Dict[str, str]:
        """Gets credentials override by merging service overrides and user overrides"""
        credentials = {}

        if self.auth_type == DatasourceDtoAuthType.OAUTH.value:
            credentials = load_oauth_credentials()

        if self.auth_type in (
            DatasourceDtoAuthType.AWSIAMROLE.value,
            DatasourceDtoAuthType.AWSIAMROLEWITHUSERNAME.value,
        ):
            credentials = load_aws_credentials(
                os.getenv(AWS_SHARED_CREDENTIALS_FILE, ""),
                getattr(self._config_override, "profile", ""),
            )

        credentials.update(self._config_override.credential())
        return credentials

    def update(self, config: DatasourceConfig) -> None:
        """Store configuration override for future query calls.

        Args:
            config: specific datasource config class
        """
        self._config_override = config

    def reset_config(self) -> None:
        """Reset the configuration override."""
        self._config_override = Config()


@attr.s
class TabularDatasource(Datasource):
    """Represents a tabular type datasource."""

    def query(self, query: str) -> Result:
        """Execute a query against the datasource.

        Args:
            query: SQL statement to execute

        Returns:
            Result entity wrapping dataframe
        """
        return self.client.execute(
            self.identifier,
            query,
            config=self._config_override.config(),
            credential=self._get_credential_override(),
        )


@attr.s
class ObjectStoreDatasource(Datasource):
    """Represents a object store type datasource."""

    def Object(self, key: str) -> _Object:  # pylint: disable=invalid-name
        """Return an object with given key and datasource client."""
        return _Object(datasource=self, key=key)

    def list_objects(self, prefix: str = "", page_size: int = 1000) -> List[_Object]:
        """List objects in the object store datasource.

        Args:
            prefix: optional prefix to filter objects
            page_size: optional number of objects to fetch

        Returns:
            List of objects
        """
        keys = self.client.list_keys(
            self.identifier,
            prefix,
            page_size,
            config=self._config_override.config(),
            credential=self._get_credential_override(),
        )
        return [
            _Object(
                datasource=self,
                key=key,
            )
            for key in keys
        ]

    def get_key_url(self, object_key: str, is_read_write: bool = False) -> str:
        """Get a signed URL for the given key.

        Args:
            object_key: unique identifier of object to get signed URL for.
            is_read_write: whether the URL should allow writes or not.

        Returns:
            Signed URL for given key
        """
        return cast(
            str,
            self.client.get_key_url(
                self.identifier,
                object_key,
                is_read_write,
                config=self._config_override.config(),
                credential=self._get_credential_override(),
            ),
        )

    def get(self, object_key: str) -> bytes:
        """Get object content as bytes.

        Args:
            object_key: unique key of object

        Returns:
            object content as bytes
        """
        return self.Object(object_key).get()

    def download_file(self, object_key: str, filename: str) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exists.

        Args:
            object_key: unique key of object
            filename: path of file to write content to.
        """
        self.Object(object_key).download_file(filename)

    def download(self, object_key: str, filename: str, max_workers: int = MAX_WORKERS) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exists.

        Args:
            object_key: unique key of object
            filename: path of file to write content to
            max_workers: max parallelism for high speed download
        """
        self.Object(object_key).download(filename, max_workers)

    def download_fileobj(self, object_key: str, fileobj: Any) -> None:
        """Download object content to file like object.

        Args:
            object_key: unique key of object
            fileobj: A file-like object to download into.
                At a minimum, it must implement the write method and must accept bytes.
        """
        self.Object(object_key).download_fileobj(fileobj)

    def put(self, object_key: str, content: bytes) -> None:
        """Upload content to object at given key.

        Args:
            object_key: unique key of object
            content: bytes content
        """
        self.Object(object_key).put(content)

    def upload_file(self, object_key: str, filename: str) -> None:
        """Upload content of file at filename to object at given key.

        Args:
            object_key: unique key of object
            filename: path of file to upload.
        """
        self.Object(object_key).upload_file(filename)

    def upload_fileobj(self, object_key: str, fileobj: Any) -> None:
        """Upload content of file like object to object at given key.

        Args:
            object_key: unique key of object
            fileobj: A file-like object to upload from.
                At a minimum, it must implement the read method and must return bytes.
        """
        self.Object(object_key).upload_fileobj(fileobj)


@attr.s
class BoardingPass:
    """Represent a query request to the Datasource Proxy service."""

    # pylint: disable=too-few-public-methods

    datasource_id: str = attr.ib()
    query: str = attr.ib()
    config: Dict[str, str] = attr.ib()
    credential: Dict[str, str] = attr.ib()

    def to_json(self) -> str:
        """Serialize self to JSON."""
        return json.dumps(
            {
                "datasourceId": self.datasource_id,
                "sqlQuery": self.query,
                "configOverwrites": self.config,
                "credentialOverwrites": self.credential,
            }
        )


@attr.s
class DataSourceClient:
    """API client and bindings."""

    domino: AuthenticatedClient = attr.ib(init=False, repr=False)
    proxy: flight.FlightClient = attr.ib(init=False, repr=False)
    proxy_http: AuthenticatedClient = attr.ib(init=False, repr=False)

    api_key: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_USER_API_KEY))
    token_file: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_TOKEN_FILE))
    token_url: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_API_PROXY))
    token: Optional[str] = attr.ib(default=None)

    def __attrs_post_init__(self):
        flight_host = os.getenv(DOMINO_DATASOURCE_PROXY_FLIGHT_HOST)
        domino_host = os.getenv(
            DOMINO_API_PROXY, os.getenv(DOMINO_API_HOST, os.getenv(DOMINO_USER_HOST, ""))
        )
        proxy_host = os.getenv(DOMINO_DATASOURCE_PROXY_HOST, "")

        client_source = os.getenv(DOMINO_CLIENT_SOURCE, "Python")
        run_id = os.getenv(DOMINO_RUN_ID, "")

        logger.info(
            "initializing datasource client with hosts",
            flight_host=flight_host,
            domino_host=domino_host,
            proxy_host=proxy_host,
        )

        self._set_proxy()
        self.proxy_http = ProxyClient(
            base_url=proxy_host,
            api_key=self.api_key,
            client_source=client_source,
            run_id=run_id,
            token_file=self.token_file,
            token_url=self.token_url,
            token=self.token,
            timeout=httpx.Timeout(5.0),
            verify_ssl=True,
        )
        self.domino = AuthenticatedClient(
            base_url=f"{domino_host}/v4",
            api_key=self.api_key,
            token_file=self.token_file,
            token_url=self.token_url,
            token=self.token,
            headers=ACCEPT_HEADERS,
            timeout=httpx.Timeout(20.0),
            verify_ssl=True,
        )

    def _set_proxy(self):
        client_source = os.getenv(DOMINO_CLIENT_SOURCE, "Python")
        flight_host = os.getenv(DOMINO_DATASOURCE_PROXY_FLIGHT_HOST)
        run_id = os.getenv(DOMINO_RUN_ID, "")
        self.proxy = flight.FlightClient(
            flight_host,
            middleware=[
                AuthMiddlewareFactory(
                    self.api_key,
                    self.token_file,
                    self.token_url,
                    self.token,
                ),
                MetaMiddlewareFactory(client_source=client_source, run_id=run_id),
            ],
        )

    def get_datasource(self, name: str) -> Datasource:
        """Fetch a datasource by name.

        Args:
            name: unique identifier of a datasource

        Returns:
            Datasource entity with given name

        Raises:
            Exception: If the response from Domino is not 200
        """
        logger.info("get_datasource", datasource_name=name)

        run_id = os.getenv(DOMINO_RUN_ID)
        response = get_datasource_by_name.sync_detailed(
            name,
            client=self.domino.with_auth_headers(),
            run_id=run_id,
        )
        if response.status_code == 200:
            datasource_dto = cast(DatasourceDto, response.parsed)
            datasource_klass = cast(
                Datasource, find_datasource_klass(datasource_dto.data_source_type)
            )
            return datasource_klass.from_dto(self, datasource_dto)

        error_response = cast(ErrorResponse, response.parsed)
        message = (
            f"Received unexpected response while getting data source: {response}"
            if error_response is None
            else error_response.message
        )
        logger.exception(message)
        raise Exception(message)

    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
    def list_keys(
        self,
        datasource_id: str,
        prefix: str,
        page_size: int,
        config: Dict[str, str],
        credential: Dict[str, str],
    ) -> List[str]:
        """List keys in a datasource.

        Args:
            datasource_id: unique identifier of a datasource
            prefix: prefix to filter keys with
            page_size: number of objects to return
            config: overwrite configuration dictionary
            credential: overwrite credential dictionary

        Returns:
            List of keys as string

        Raises:
            Exception: if the response from the Proxy is not 200
            UnauthenticatedError: if the request has invalid authentication
        """
        logger.info("list_keys", datasource_id=datasource_id, prefix=prefix, page_size=page_size)

        response = list_keys.sync_detailed(
            client=self.proxy_http.with_auth_headers(),
            body=ListRequest(
                datasource_id=datasource_id,
                prefix=prefix,
                page_size=page_size,
                config_overwrites=APIConfig.from_dict(config),
                credential_overwrites=APIConfig.from_dict(credential),
            ),
        )

        if response.status_code == 200:
            return cast(List[str], response.parsed)
        if response.status_code == 401:
            raise UnauthenticatedError

        error = cast(ProxyErrorResponse, response.parsed)
        logger.exception(error)
        raise Exception(f"{error.error_type}: {error.raw_error}")

    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
    def get_key_url(  # pylint: disable=too-many-arguments
        self,
        datasource_id: str,
        object_key: str,
        is_read_write: bool,
        config: Dict[str, str],
        credential: Dict[str, str],
    ) -> str:
        """Request a signed URL for a given datasource and object key.

        Args:
            datasource_id: unique identifier of a datasource
            object_key: unique identifier of key to retrieve
            is_read_write: whether the signed URL allows write or not.
            config: overwrite configuration dictionary
            credential: overwrite credential dictionary

        Returns:
            Signed URL of the requested object.

        Raises:
            Exception: if the response from the Proxy is not 200
            UnauthenticatedError: if the request has invalid authentication
        """
        logger.info("get_key_url", datasource_id=datasource_id, object_key=object_key)

        response = get_key_url.sync_detailed(
            client=self.proxy_http.with_auth_headers(),
            body=KeyRequest(
                datasource_id=datasource_id,
                object_key=object_key,
                is_read_write=is_read_write,
                config_overwrites=APIConfig.from_dict(config),
                credential_overwrites=APIConfig.from_dict(credential),
            ),
        )

        if response.status_code == 200:
            return cast(str, response.parsed)
        if response.status_code == 401:
            raise UnauthenticatedError

        error = cast(ProxyErrorResponse, response.parsed)
        logger.exception(error)
        raise Exception(f"{error.error_type}: {error.raw_error}")

    def _log_metric(
        self,
        datasource_type: str,
        bytes_processed: int,
        is_read_write: bool,
    ) -> None:
        """Log metric about file size being processed.

        Args:
            datasource_type: type of datasource
            bytes_processed: count of read or written bytes
            is_read_write: whether used for read or write
        """
        mode = LogMetricM.WRITE if is_read_write else LogMetricM.READ
        type_map = {
            DatasourceDtoDataSourceType.ADLSCONFIG.value: LogMetricT.ADLSCONFIG,
            DatasourceDtoDataSourceType.GCSCONFIG.value: LogMetricT.GCSCONFIG,
            DatasourceDtoDataSourceType.GENERICS3CONFIG.value: LogMetricT.GENERICS3CONFIG,  # noqa
            DatasourceDtoDataSourceType.S3CONFIG.value: LogMetricT.S3CONFIG,
        }
        type_ = type_map.get(datasource_type)
        if not type_:
            return

        try:
            log_metric.sync_detailed(
                client=self.proxy_http.with_auth_headers(),
                t=type_,
                b=bytes_processed,
                m=mode,
            )
        except Exception as exc:  # pylint: disable=broad-except
            logger.exception(exc)

    def execute(
        self,
        datasource_id: str,
        query: str,
        config: Dict[str, str],
        credential: Dict[str, str],
    ) -> Result:
        """Execute a given query against a datasource.

        Args:
            datasource_id: unique identifier of a datasource
            query: SQL query to execute
            config: overwrite configuration dictionary
            credential: overwrite credential dictionary

        Returns:
            Result entity encapsulating execution response

        Raises:
            DominoError: if the proxy fails to query or return data
        """
        logger.info("execute", datasource_id=datasource_id, query=query)

        try:
            self._set_proxy()
            reader = self._do_get(
                BoardingPass(
                    datasource_id=datasource_id,
                    query=query,
                    config=config,
                    credential=credential,
                ).to_json()
            )
        except (flight.FlightError, ArrowException) as exc:
            logger.exception(exc)
            raise DominoError(_unpack_flight_error(str(exc))) from None
        return Result(self, reader, query)

    @backoff.on_exception(backoff.expo, flight.FlightUnauthenticatedError, max_time=60)
    def _do_get(self, ticket: str) -> flight.FlightStreamReader:
        return self.proxy.do_get(flight.Ticket(ticket))



"""Authentication classes for HTTP and Flight clients."""

from typing import Dict, Optional

from os.path import exists

import attr
import backoff
import httpx
from httpx._config import DEFAULT_TIMEOUT_CONFIG
from pyarrow import flight

from datasource_api_client.client import Client


@backoff.on_exception(backoff.expo, httpx.HTTPStatusError, max_time=2)
@backoff.on_exception(backoff.expo, httpx.ReadTimeout, max_tries=2)
def get_jwt_token(url: str) -> str:
    """Gets a domino token from local sidecar API.

    Args:
        url: base url of sidecar container serving token API

    Returns:
        JWT as string

    Raises:
        HTTPStatusError: if the API returns an error
    """
    resp = httpx.get(f"{url}/access-token", timeout=DEFAULT_TIMEOUT_CONFIG)
    resp.raise_for_status()
    return resp.read().decode("ascii")


@attr.s(auto_attribs=True)
class AuthenticatedClient(Client):
    """A client that authenticates all requests with either the API Key or JWT."""

    api_key: Optional[str] = attr.ib()
    token_file: Optional[str] = attr.ib()
    token_url: Optional[str] = attr.ib()
    token: Optional[str] = attr.ib()

    def __attrs_post_init__(self):
        if not (self.api_key or self.token_file or self.token_url or self.token):
            raise Exception(
                "One of two authentication methods must be supplied (API Key or JWT Location)"  # noqa
            )

    def _get_auth_headers(self) -> Dict[str, str]:
        """Get auth headers with either JWT or API Key."""
        if self.token is not None:
            return {"Authorization": f"Bearer {self.token}"}
        if self.token_url is not None:
            try:
                jwt = get_jwt_token(self.token_url)
            except httpx.HTTPStatusError:
                pass
            else:
                return {"Authorization": f"Bearer {jwt}"}

        if self.token_file and exists(self.token_file):
            with open(self.token_file, encoding="ascii") as token_file:
                jwt = token_file.readline().rstrip()
            return {"Authorization": f"Bearer {jwt}"}

        if self.api_key:
            return {"X-Domino-Api-Key": self.api_key}

        return {}

    def with_auth_headers(self) -> "AuthenticatedClient":
        auth_headers = self._get_auth_headers()
        if self._client is not None:
            self._client.headers.update(auth_headers)
        if self._async_client is not None:
            self._async_client.headers.update(auth_headers)
        self._headers.update(auth_headers)
        return self


@attr.s(auto_attribs=True)
class ProxyClient(AuthenticatedClient):
    """A client that authenticates all requests but with Proxy headers."""

    client_source: Optional[str] = attr.ib()
    run_id: Optional[str] = attr.ib()

    def _get_auth_headers(self) -> Dict[str, str]:
        headers = {}
        if self.api_key:
            headers["X-Domino-Api-Key"] = self.api_key
        if self.client_source:
            headers["X-Domino-Client-Source"] = self.client_source
        if self.run_id:
            headers["X-Domino-Run-Id"] = self.run_id
        if self.token is not None:
            headers["Authorization"] = f"Bearer {self.token}"

        if self.token_url is not None:
            try:
                jwt = get_jwt_token(self.token_url)
            except httpx.HTTPStatusError:
                pass
            else:
                headers["X-Domino-Jwt"] = jwt
                return headers

        if self.token_file and exists(self.token_file):
            with open(self.token_file, encoding="ascii") as token_file:
                jwt = token_file.readline().rstrip()
            headers["X-Domino-Jwt"] = jwt

        return headers


@attr.s(auto_attribs=True)
class AuthMiddlewareFactory(flight.ClientMiddlewareFactory):
    """Middleware Factory for authenticating flight requests."""

    api_key: Optional[str] = attr.ib()
    token_file: Optional[str] = attr.ib()
    token_url: Optional[str] = attr.ib()
    token: Optional[str] = attr.ib()

    def __attrs_post_init__(self):
        if not (self.api_key or self.token_file or self.token_url or self.token):
            raise Exception(
                "One of two authentication methods must be supplied (API Key or JWT Location)"  # noqa
            )

    def start_call(self, _):
        """Called at the start of an RPC."""
        jwt = None
        if self.token is not None:
            return {"Authorization": f"Bearer {self.token}"}

        if self.token_url is not None:
            try:
                jwt = get_jwt_token(self.token_url)
            except httpx.HTTPStatusError:
                pass
            else:
                return AuthMiddleware(self.api_key, jwt)

        if self.token_file is not None and exists(self.token_file):
            with open(self.token_file, encoding="ascii") as token_file:
                jwt = token_file.readline().rstrip()

        return AuthMiddleware(self.api_key, jwt)


@attr.s(auto_attribs=True)
class AuthMiddleware(flight.ClientMiddleware):
    """Middleware for authenticating flight requests."""

    api_key: Optional[str] = attr.ib()
    jwt: Optional[str] = attr.ib()

    def call_completed(self, _):
        """No implementation. TODO logging."""

    def received_headers(self, _):
        """No implementation."""

    def sending_headers(self):
        """Return authentication headers."""
        headers = {}

        if self.api_key is not None:
            headers["x-domino-api-key"] = self.api_key
        if self.jwt is not None:
            headers["x-domino-jwt"] = self.jwt

        return headers



from typing import List, Mapping, Optional

import os
from dataclasses import dataclass, field

import pandas as pd


@dataclass
class TrainingSet:
    """
    A Training Set.

    Args:
        name: Unique name of the TrainingSet.
        description: Description of the TrainingSet.
        meta: User defined metadata.
    """

    name: str
    project_id: str
    description: Optional[str] = None
    meta: Mapping[str, str] = field(default_factory=map)


@dataclass
class MonitoringMeta:
    """
    Monitoring Meta.

    For more details about the parameters, refer to :class:`.TrainingSetVersion`.

    Args:
        timestamp_columns: Timestamp columns.
        categorical_columns: Categorical columns.
        ordinal_columns: Ordinal columns. Currently, ordinal columns are skipped by the
            Model Monitor.
    """

    timestamp_columns: List[str] = field(default_factory=list)
    categorical_columns: List[str] = field(default_factory=list)
    ordinal_columns: List[str] = field(default_factory=list)


@dataclass
class TrainingSetVersion:
    """
    A Training Set Version.

    Any columns that are not inside ``key_columns``, ``exclude_columns``, \
        ``MonitoringMeta.categorical_columns``, ``MonitoringMeta.timestamp_columns``, \
        or ``MonitoringMeta.ordinal_columns`` are automatically marked as numerical \
        columns.

    Args:
        number: The TrainingSetVersion number.
        training_set_name: Name of the TrainingSet this version belongs to.
        description: Description of this version.
        key_columns: Row identifier columns.
        target_columns:
            Prediction columns.

            * For **classifications models**, this must be a **categorical** column. \
                Be sure to also include this column in \
                :py:attr:`.MonitoringMeta.categorical_columns`.

            * For **regression models**, it must be a **numerical** column.
        exclude_columns: Any columns that should be excluded.
        all_columns: Names all columns in the dataframe.
        monitoring_meta: Monitoring specific metadata.
        meta: User defined metadata
    """

    training_set_name: str
    number: int
    description: Optional[str] = None
    key_columns: List[str] = field(default_factory=list)
    target_columns: List[str] = field(default_factory=list)
    exclude_columns: List[str] = field(default_factory=list)
    all_columns: List[str] = field(default_factory=list)
    monitoring_meta: MonitoringMeta = field(default_factory=MonitoringMeta)
    meta: Mapping[str, str] = field(default_factory=map)
    path: Optional[str] = None
    container_path: Optional[str] = None
    pending: bool = True

    @property
    def absolute_container_path(self):
        root = os.getenv("DOMINO_TRAINING_SET_PATH", "/var/opt/domino/trainingset/project")

        return f"{root}/{self.container_path}"

    def load_training_pandas(self) -> pd.DataFrame:
        """Get a pandas dataframe for training.

        Dataframe does not include key_columns and exclude_columns.
        """

        df = self.load_raw_pandas()

        for c in self.key_columns + self.exclude_columns:
            if c in df.columns:
                del df[c]

        return df

    def load_raw_pandas(self) -> pd.DataFrame:
        """Get the raw dataframe."""

        files = [
            os.path.join(self.absolute_container_path)
            for f in os.listdir(self.absolute_container_path)
            if f.endswith(".parquet")
        ]

        return pd.concat([pd.read_parquet(f) for f in files])



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.auth_type import AuthType

if TYPE_CHECKING:
    from ..models.auth_config_fields import AuthConfigFields
    from ..models.auth_config_meta import AuthConfigMeta


T = TypeVar("T", bound="AuthConfig")


@_attrs_define
class AuthConfig:
    """
    Attributes:
        auth_type (AuthType):
        fields (AuthConfigFields):
        meta (AuthConfigMeta):
    """

    auth_type: AuthType
    fields: "AuthConfigFields"
    meta: "AuthConfigMeta"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        auth_type = self.auth_type.value

        fields = self.fields.to_dict()

        meta = self.meta.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "authType": auth_type,
                "fields": fields,
                "meta": meta,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.auth_config_fields import AuthConfigFields
        from ..models.auth_config_meta import AuthConfigMeta

        d = src_dict.copy()
        auth_type = AuthType(d.pop("authType"))

        fields = AuthConfigFields.from_dict(d.pop("fields"))

        meta = AuthConfigMeta.from_dict(d.pop("meta"))

        auth_config = cls(
            auth_type=auth_type,
            fields=fields,
            meta=meta,
        )

        auth_config.additional_properties = d
        return auth_config

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Domino TrainingSet client library."""

from typing import List, Mapping, Optional

import json
import os
import re
import shutil
from stat import S_IRGRP, S_IROTH, S_IRUSR, S_IWUSR, S_IXGRP, S_IXOTH, S_IXUSR

import pandas as pd

from training_set_api_client.api.default import (
    delete_training_set_name,
    delete_training_set_name_number,
    get_training_set_name,
    get_training_set_name_number,
    post_find,
    post_training_set_name,
    post_version_find,
    put_training_set_name,
    put_training_set_name_number,
)
from training_set_api_client.models import (
    CreateTrainingSetVersionRequest,
    CreateTrainingSetVersionRequestMeta,
    MonitoringMeta,
    TrainingSet,
    TrainingSetFilter,
    TrainingSetFilterMeta,
    TrainingSetVersion,
    TrainingSetVersionFilter,
    TrainingSetVersionFilterMeta,
    TrainingSetVersionFilterTrainingSetMeta,
    UpdateTrainingSetRequest,
    UpdateTrainingSetRequestMeta,
    UpdateTrainingSetVersionRequest,
    UpdateTrainingSetVersionRequestMeta,
)
from training_set_api_client.types import Response

from ..auth import AuthenticatedClient
from ..training_sets import model

_trainingset_name_pat = re.compile("^[-A-Za-z0-9_]+$")


class ServerException(Exception):
    """This exception is raised when the TrainingSet server rejects a request."""

    def __init__(self, message: str, server_msg: str):
        self.message = message
        self.server_msg = server_msg


class SchemaMismatchException(Exception):
    """This exception is raised when the TrainingSet data columns do not match the metadata."""


def get_training_set(name: str) -> model.TrainingSet:
    """Get a TrainingSet by name.

    Args:
        name: Name of the training set.

    Returns:
        The TrainingSet, if found.

    """

    _validate_trainingset_name(name)

    with _get_client() as client:
        response = get_training_set_name.sync_detailed(
            client=client,
            training_set_name=name,
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not get TrainingSet")

        return _to_TrainingSet(response.parsed)


def list_training_sets(
    meta: Optional[Mapping[str, str]] = None,
    asc: bool = True,
    offset: int = 0,
    limit: int = 10000,
) -> List[model.TrainingSet]:
    """Query training sets.

    Args:
        meta: Metadata key-value pairs to match.
        asc: Sort order by creation time, 1 for ascending -1 for descending.
        offset: Offset
        limit: Limit

    Returns:
        A list of matching TrainingSets.
    """

    if meta is None:
        meta = {}

    project_id = _get_project_id()

    with _get_client() as client:
        response = post_find.sync_detailed(
            client=client,
            body=TrainingSetFilter(
                project_id=project_id,
                meta=TrainingSetFilterMeta.from_dict(meta),
            ),
            offset=offset,
            limit=limit,
            asc=asc,
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not list TrainingSets")

        return [_to_TrainingSet(ts) for ts in response.parsed]


def update_training_set(
    updated: model.TrainingSet,
) -> model.TrainingSet:
    """Update a TrainingSet.

    Args:
        updated: Updated TrainingSet.

    Returns:
        The updated TrainingSet from the server.
    """

    _validate_trainingset_name(updated.name)

    with _get_client() as client:
        response = put_training_set_name.sync_detailed(
            training_set_name=updated.name,
            client=client,
            body=UpdateTrainingSetRequest(
                meta=UpdateTrainingSetRequestMeta.from_dict(updated.meta),
                description=updated.description,
            ),
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not update TrainingSets")

        return _to_TrainingSet(response.parsed)


def delete_training_set(name: str) -> bool:
    """Delete a TrainingSet.

    **Note:** This deletes the TrainingSet only if it has no versions.

    Args:
        name: Name of the TrainingSet.

    Returns:
        True if TrainingSet was deleted.
    """

    _validate_trainingset_name(name)

    with _get_client() as client:
        response = delete_training_set_name.sync_detailed(training_set_name=name, client=client)

        if response.status_code != 200:
            _raise_response_exn(response, "could not delete TrainingSet")

        return True


def create_training_set_version(
    training_set_name: str,
    df: pd.DataFrame,
    description: Optional[str] = None,
    key_columns: Optional[List[str]] = None,
    target_columns: Optional[List[str]] = None,
    exclude_columns: Optional[List[str]] = None,
    monitoring_meta: Optional[model.MonitoringMeta] = None,
    meta: Optional[Mapping[str, str]] = None,
    **kwargs,
) -> model.TrainingSetVersion:
    """Create a TrainingSetVersion.

    Args:
        training_set_name: Name of the TrainingSet this version belongs to. ``training_set_name`` must
            be a string containing only alphanumeric characters in the basic Latin alphabet
            including dash and underscore: `[-A-Za-z_]`.
        df: A DataFrame holding the data.
        description: Description of this version.
        key_columns: Names of columns that represent IDs for retrieving features.
        target_columns: Target variables for prediction.
        exclude_columns: Columns to exclude when generating the training DataFrame.
        monitoring_meta: Monitoring specific metadata.
        meta: User defined metadata.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        The created TrainingSetVersion
    """

    if key_columns is None:
        key_columns = []

    if target_columns is None:
        target_columns = []

    if exclude_columns is None:
        exclude_columns = []

    if monitoring_meta is None:
        monitoring_meta = model.MonitoringMeta()

    if meta is None:
        meta = {}

    all_columns = list(df.columns)

    _validate_trainingset_name(training_set_name)

    _check_columns(
        all_columns,
        key_columns
        + target_columns
        + exclude_columns
        + monitoring_meta.timestamp_columns
        + monitoring_meta.categorical_columns
        + monitoring_meta.ordinal_columns,
    )

    project_id = _get_project_id()

    with _get_client() as client:
        response = post_training_set_name.sync_detailed(
            client=client,
            training_set_name=training_set_name,
            body=CreateTrainingSetVersionRequest(
                project_id=project_id,
                key_columns=key_columns,
                target_columns=target_columns,
                exclude_columns=exclude_columns,
                all_columns=all_columns,
                monitoring_meta=MonitoringMeta(
                    timestamp_columns=monitoring_meta.timestamp_columns,
                    categorical_columns=monitoring_meta.categorical_columns,
                    ordinal_columns=monitoring_meta.ordinal_columns,
                ),
                meta=CreateTrainingSetVersionRequestMeta.from_dict(meta),
                description=description,
            ),
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not create Training Set version")

        tsv = _to_TrainingSetVersion(response.parsed)

        os.makedirs(tsv.absolute_container_path)
        df.to_parquet(os.path.join(tsv.absolute_container_path, "data.parquet"))
        os.chmod(
            tsv.absolute_container_path, S_IRUSR | S_IRGRP | S_IROTH | S_IXUSR | S_IXGRP | S_IXOTH
        )

        tsv.pending = False
        return update_training_set_version(tsv)


def get_training_set_version(training_set_name: str, number: int) -> model.TrainingSetVersion:
    """Gets a TrainingSetVersion by version number.

    Args:
        training_set_name: Name of the TrainingSet.
        number: Version number.

    Returns:
        The requested TrainingSetVersion.
    """

    _validate_trainingset_name(training_set_name)

    with _get_client() as client:
        response = get_training_set_name_number.sync_detailed(
            client=client,
            training_set_name=training_set_name,
            number=number,
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not get TrainingSetVersion")

        return _to_TrainingSetVersion(response.parsed)


def update_training_set_version(version: model.TrainingSetVersion) -> model.TrainingSetVersion:
    """Updates this TrainingSetVersion.

    Args:
        version: TrainingSetVersion to update.

    Returns:
        The updated TrainingSetVersion from the server.
    """

    _validate_trainingset_name(version.training_set_name)

    with _get_client() as client:
        response = put_training_set_name_number.sync_detailed(
            training_set_name=version.training_set_name,
            number=version.number,
            client=client,
            body=UpdateTrainingSetVersionRequest(
                key_columns=version.key_columns,
                target_columns=version.target_columns,
                exclude_columns=version.exclude_columns,
                monitoring_meta=MonitoringMeta(
                    timestamp_columns=version.monitoring_meta.timestamp_columns,
                    categorical_columns=version.monitoring_meta.categorical_columns,
                    ordinal_columns=version.monitoring_meta.ordinal_columns,
                ),
                meta=UpdateTrainingSetVersionRequestMeta.from_dict(version.meta),
                pending=version.pending,
                description=version.description,
            ),
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not update TrainingSetVersion")

        return _to_TrainingSetVersion(response.parsed)


def delete_training_set_version(training_set_name: str, number: int) -> bool:
    """Deletes a TrainingSetVersion.

    Args:
        training_set_name: Name of the TrainingSet.
        number: TrainingSetVersion number.

    Returns:
        True if TrainingSetVersion was deleted.
    """

    _validate_trainingset_name(training_set_name)

    tsv = get_training_set_version(training_set_name, number)

    with _get_client() as client:
        response = delete_training_set_name_number.sync_detailed(
            training_set_name=training_set_name,
            number=number,
            client=client,
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not delete TrainingSetVersion")

        stat = os.stat(tsv.absolute_container_path)
        os.chmod(tsv.absolute_container_path, stat.st_mode | S_IWUSR)
        shutil.rmtree(tsv.absolute_container_path)

        return True


def list_training_set_versions(
    meta: Optional[Mapping[str, str]] = None,
    training_set_name: Optional[str] = None,
    training_set_meta: Optional[Mapping[str, str]] = None,
    asc: bool = True,
    offset: int = 0,
    limit: int = 10000,
) -> List[model.TrainingSetVersion]:
    """List training sets.

    Args:
        meta: Version metadata.
        training_set_name: Training set name.
        training_set_meta: Training set meta data.
        asc: Sort order by creation time, 1 for ascending -1 for descending.
        offset: Offset.
        limit: Limit.

    Returns:
        A list of matching TrainingSetVersions.
    """

    if meta is None:
        meta = {}

    if training_set_meta is None:
        training_set_meta = {}

    project_id = _get_project_id()

    with _get_client() as client:
        response = post_version_find.sync_detailed(
            client=client,
            body=TrainingSetVersionFilter(
                training_set_meta=TrainingSetVersionFilterTrainingSetMeta.from_dict(
                    training_set_meta,
                ),
                meta=TrainingSetVersionFilterMeta.from_dict(meta),
                project_id=project_id,
                training_set_name=training_set_name,
            ),
            offset=offset,
            limit=limit,
            asc=asc,
        )

        if response.status_code != 200:
            _raise_response_exn(response, "could not find TrainingSetVersion")

        return [_to_TrainingSetVersion(tsv) for tsv in response.parsed]


def _get_client() -> AuthenticatedClient:
    domino_host = os.getenv("DOMINO_API_HOST", os.getenv("DOMINO_USER_HOST"))
    api_key = os.getenv("DOMINO_USER_API_KEY")
    token_file = os.getenv("DOMINO_TOKEN_FILE")
    token_url = os.getenv("DOMINO_API_PROXY")

    return AuthenticatedClient(
        base_url=f"{domino_host}/trainingset",
        api_key=api_key,
        token_file=token_file,
        token_url=token_url,
        token=None,
    ).with_auth_headers()


def _to_TrainingSet(ts: TrainingSet) -> model.TrainingSet:
    return model.TrainingSet(
        name=ts.name,
        description=ts.description,
        meta=ts.meta.to_dict(),
        project_id=ts.project_id,
    )


def _to_TrainingSetVersion(tsv: TrainingSetVersion) -> model.TrainingSetVersion:
    return model.TrainingSetVersion(
        training_set_name=tsv.training_set_name,
        number=tsv.number,
        description=tsv.description,
        key_columns=tsv.key_columns,
        target_columns=tsv.target_columns,
        exclude_columns=tsv.exclude_columns,
        all_columns=tsv.all_columns,
        monitoring_meta=model.MonitoringMeta(
            timestamp_columns=tsv.monitoring_meta.timestamp_columns,
            categorical_columns=tsv.monitoring_meta.categorical_columns,
            ordinal_columns=tsv.monitoring_meta.ordinal_columns,
        ),
        meta=tsv.meta.to_dict(),
        path=tsv.path,
        container_path=tsv.container_path,
        pending=tsv.pending,
    )


def _raise_response_exn(response: Response, msg: str):
    try:
        response_json = json.loads(response.content.decode("utf8"))
        server_msg = response_json.get("errors")
    except Exception:
        server_msg = None

    raise ServerException(msg, server_msg)


def _check_columns(all_columns: [str], expected_columns: [str]):
    diff = set(expected_columns) - set(all_columns)
    if diff:
        raise SchemaMismatchException(f"DataFrame missing columns: {diff}")


def _get_project_id() -> Optional[str]:
    return os.getenv("DOMINO_PROJECT_ID")


def _validate_trainingset_name(name: str):
    if _trainingset_name_pat.match(name) is None:
        raise ValueError(f"bad TrainingSet name '{name}'")



"""Dedicated Feature Store Logging module."""

import os
import sys

from loguru import logger as _logger


def get_feature_store_logger():
    """Configure and return a logger."""
    _logger.configure(
        handlers=[
            {
                "format": "[{time}] {message}",
                "sink": sys.stdout,
                "enqueue": True,
            },
        ],
        extra={
            "ip": os.getenv("DOMINO_NODE_IP"),
            "project": os.getenv("DOMINO_PROJECT_NAME"),
            "run_id": os.getenv("DOMINO_RUN_ID"),
            "user": os.getenv("DOMINO_STARTING_USERNAME"),
        },
    )
    return _logger


logger = get_feature_store_logger()



"""Datasource module."""

from typing import Any, List, Optional

import os
from os.path import exists

import attr
import backoff
import httpx
import urllib3

import domino_data.configuration_gen
from domino_data.data_sources import DataSourceClient, ObjectStoreDatasource

from .auth import AuthenticatedClient, get_jwt_token
from .configuration_gen import Config, DatasetConfig
from .logging import logger
from .transfer import MAX_WORKERS, BlobTransfer

ACCEPT_HEADERS = {"Accept": "application/json"}

DOMINO_API_HOST = "DOMINO_API_HOST"
DOMINO_API_PROXY = "DOMINO_API_PROXY"
DOMINO_USER_API_KEY = "DOMINO_USER_API_KEY"
DOMINO_USER_HOST = "DOMINO_USER_HOST"
DOMINO_TOKEN_FILE = "DOMINO_TOKEN_FILE"


def __getattr__(name: str) -> Any:
    if name.endswith("Config"):
        return getattr(domino_data.configuration_gen, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


def __dir__() -> List[str]:
    confs = filter(lambda x: x.endswith("Config"), dir(domino_data.configuration_gen))
    return list(globals().keys()) + list(confs)


class DominoError(Exception):
    """Base exception for known errors."""


class UnauthenticatedError(DominoError):
    """To handle exponential backoff."""


@attr.s
class _File:
    """Represents a file in a dataset."""

    dataset: "Dataset" = attr.ib(repr=False)
    name: str = attr.ib(repr=False)

    def http(self) -> httpx.Client:
        """Get dataset http client."""
        return self.dataset.http()

    def pool_manager(self) -> urllib3.PoolManager:
        """Get dataset http pool manager."""
        return self.dataset.pool_manager()

    def get(self) -> bytes:
        """Get object content as bytes."""
        url = self.dataset.get_file_url(self.name)
        headers = self._get_headers()

        res = self.http().get(url, headers=headers)
        res.raise_for_status()

        return res.content

    def download_file(self, filename: str) -> None:
        """Download object content to file located at filename.

        The file will be created if it does not exist.

        Args:
            filename: path of file to write content to.
        """
        url = self.dataset.get_file_url(self.name)
        headers = self._get_headers()

        content_size = 0
        with (
            self.http().stream("GET", url, headers=headers) as stream,
            open(filename, "wb") as file,
        ):
            for data in stream.iter_bytes():
                content_size += len(data)
                file.write(data)

    def download(self, filename: str, max_workers: int = MAX_WORKERS) -> None:
        """Download object content to file with multithreaded support.

        The file will be created if it does not exist. File will be overwritten if it exists.

        Args:
            filename: path of file to write content to
            max_workers: max parallelism for high speed download
        """
        url = self.dataset.get_file_url(self.name)
        headers = self._get_headers()
        with open(filename, "wb") as file:
            BlobTransfer(
                url, file, headers=headers, max_workers=max_workers, http=self.pool_manager()
            )

    def download_fileobj(self, fileobj: Any) -> None:
        """Download object content to file like object.

        Args:
            fileobj: A file-like object to download into.
                At a minimum, it must implement the write method and must accept bytes.
        """
        url = self.dataset.get_file_url(self.name)
        headers = self._get_headers()
        content_size = 0
        with self.http().stream("GET", url, headers=headers) as stream:
            for data in stream.iter_bytes():
                content_size += len(data)
                fileobj.write(data)

    def _get_headers(self) -> dict:
        headers = {}

        if self.dataset.client.token_url is not None:
            try:
                jwt = get_jwt_token(self.dataset.client.token_url)
            except httpx.HTTPStatusError:
                pass
            else:
                headers = {"Authorization": f"Bearer {jwt}"}

        if self.dataset.client.token_file and exists(self.dataset.client.token_file):
            with open(self.dataset.client.token_file, encoding="ascii") as token_file:
                jwt = token_file.readline().rstrip()
            headers = {"Authorization": f"Bearer {jwt}"}

        if self.dataset.client.api_key:
            headers = {"X-Domino-Api-Key": self.dataset.client.api_key}

        if self.dataset.client.token is not None:
            headers = {"Authorization": f"Bearer {self.dataset.client.token}"}

        return headers


@attr.s
class Dataset:
    """Represents a Domino dataset."""

    # pylint: disable=too-many-instance-attributes

    client: "DatasetClient" = attr.ib(repr=False)
    datasource: ObjectStoreDatasource = attr.ib(repr=False)

    def http(self) -> httpx.Client:
        """Singleton http client built for the dataset."""
        return self.datasource.http()

    def pool_manager(self) -> urllib3.PoolManager:
        """Urllib3 pool manager for range downloads."""
        return urllib3.PoolManager()

    def update(self, config: DatasetConfig) -> None:
        """Store configuration override for future query calls.

        Args:
            config: dataset config class
        """
        self.datasource._config_override = config

    def reset_config(self) -> None:
        """Reset the configuration override."""
        self.datasource._config_override = Config()

    def File(self, file_name: str) -> _File:  # pylint: disable=invalid-name
        """Return a file with given name and dataset client."""
        return _File(dataset=self, name=file_name)

    def list_files(self, prefix: str = "", page_size: int = 1000) -> List[_File]:
        """List files in the dataset.

        Args:
            prefix: optional prefix to filter files
            page_size: optional number of files to fetch

        Returns:
            List of files
        """
        objects = self.datasource.list_objects(prefix=prefix, page_size=page_size)

        return [_File(dataset=self, name=obj.key) for obj in objects]

    def get_file_url(self, file_name: str) -> str:
        """Get a signed URL for the given key.

        Args:
            file_name: name of the file to get URL for.

        Returns:
            URL for the given file
        """
        return self.datasource.get_key_url(object_key=file_name, is_read_write=False)

    def get(self, file_name: str) -> bytes:
        """Get file contents as bytes.

        Args:
            file_name: name of the file

        Returns:
            file contents as bytes
        """
        return self.File(file_name).get()

    def download_file(self, dataset_file_name: str, local_file_name: str) -> None:
        """Download file content to file located at local_file_name.

        The file will be created if it does not exist.

        Args:
            dataset_file_name: name of the file in the dataset to download.
            local_file_name: path of file to write content to.
        """
        self.File(dataset_file_name).download_file(local_file_name)

    def download(
        self, dataset_file_name: str, local_file_name: str, max_workers: int = MAX_WORKERS
    ) -> None:
        """Download file content to file located at filename.

        The file will be created if it does not exist.

        Args:
            dataset_file_name: name of the file in the dataset to download.
            local_file_name: path of file to write content to
            max_workers: max parallelism for high speed download
        """
        self.File(dataset_file_name).download(local_file_name, max_workers)

    def download_fileobj(self, dataset_file_name: str, fileobj: Any) -> None:
        """Download file contents to file like object.

        Args:
            dataset_file_name: name of the file in the dataset to download.
            fileobj: A file-like object to download into.
                At a minimum, it must implement the write method and must accept bytes.
        """
        self.File(dataset_file_name).download_fileobj(fileobj)


@attr.s
class DatasetClient:
    """API client and bindings."""

    api_key: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_USER_API_KEY))
    token_file: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_TOKEN_FILE))
    token_url: Optional[str] = attr.ib(factory=lambda: os.getenv(DOMINO_API_PROXY))
    token: Optional[str] = attr.ib(default=None)

    domino: AuthenticatedClient = attr.ib(init=False, repr=False)
    datasource_client = attr.ib(init=False, repr=False)

    def __attrs_post_init__(self):
        domino_host = os.getenv(
            DOMINO_API_PROXY, os.getenv(DOMINO_API_HOST, os.getenv(DOMINO_USER_HOST, ""))
        )

        logger.info(
            "initializing dataset client with host",
            domino_host=domino_host,
        )

        self.domino = AuthenticatedClient(
            base_url=f"{domino_host}/v4",
            api_key=self.api_key,
            token_file=self.token_file,
            token_url=self.token_url,
            token=self.token,
            headers=ACCEPT_HEADERS,
            timeout=httpx.Timeout(20.0),
            verify_ssl=True,
        )

        self.datasource_client = DataSourceClient(
            api_key=self.api_key,
            token_file=self.token_file,
            token_url=self.token_url,
            token=self.token,
        )

    def get_dataset(self, name: str) -> Dataset:
        """Fetch a dataset by name.

        Args:
            name: unique name of a dataset

        Returns:
            Dataset entity with given name

        Raises:
            Exception: If the response from Domino is not 200
        """
        logger.info("get_dataset", dataset_name=name)

        data_source = self.datasource_client.get_datasource(name)
        return Dataset(client=self, datasource=data_source)

    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
    def list_files(
        self,
        dataset_unique_name: str,
        prefix: str,
        page_size: int,
    ) -> List[str]:
        """List files in a dataset.

        Args:
            dataset_unique_name: unique name of a dataset
            prefix: optional prefix to filter files
            page_size: optional number of files to fetch

        Returns:
            List of files as string

        Raises:
            Exception: if the response from the Proxy is not 200
            UnauthenticatedError: if the request has invalid authentication
        """
        logger.info("list_files", dataset_unique_name=dataset_unique_name)

        return self.datasource_client.list_keys(
            datasource_id=dataset_unique_name,
            prefix=prefix,
            page_size=page_size,
            config={},
            credential={},
        )

    @backoff.on_exception(backoff.expo, UnauthenticatedError, max_time=60)
    def get_file_url(  # pylint: disable=too-many-arguments
        self,
        dataset_unique_name: str,
        file_name: str,
    ) -> str:
        """Request a URL for a given dataset and file.

        Args:
            dataset_unique_name: unique name of a dataset
            file_name: name of a file in the dataset

        Returns:
            URL of the requested file.

        Raises:
            Exception: if the response from the Proxy is not 200
            UnauthenticatedError: if the request has invalid authentication
        """
        logger.info("get_file_url", dataset_unique_name=dataset_unique_name, file_name=file_name)

        return self.datasource_client.get_key_url(
            datasource_id=dataset_unique_name,
            object_key=file_name,
            is_read_write=False,
            config={},
            credential={},
        )



from typing import Dict, Union

import os

HEADER_DOMINO_DATASOURCE = "X-Domino-Datasource"
HEADER_PINECONE_INDEX = "X-Domino-Pinecone-Index"


def domino_pinecone3x_init_params(datasource_name: str) -> Dict[str, Union[str, Dict[str, str]]]:
    """Wrap the parameters to initialize a Pinecone 3.x and above client

    Args:
        datasource_name: the name of the Pinecone data source

    Returns:
        A dictionary of parameters to initialize the Pinecone client
    """
    return {
        "api_key": "domino",
        "host": os.getenv("DOMINO_DATA_API_GATEWAY", "http://127.0.0.1:8766"),
        "additional_headers": {HEADER_DOMINO_DATASOURCE: datasource_name},
    }


def domino_pinecone3x_index_params(
    datasource_name: str, index_name: str
) -> Dict[str, Union[str, Dict[str, str]]]:
    """Wrap the parameters to target an index in the Pinecone 3.x and above client

    Args:
        datasource_name: the name of the Pinecone data source
        index_name: the name of the index

    Returns:
        A dictionary of parameters to target the Pinecone index for vector operations
    """
    return {
        "host": os.getenv("DOMINO_DATA_API_GATEWAY", "http://127.0.0.1:8766"),
        "additional_headers": {
            HEADER_DOMINO_DATASOURCE: datasource_name,
            HEADER_PINECONE_INDEX: index_name,
        },
    }



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.datasource_dto_auth_type import DatasourceDtoAuthType
from ..models.datasource_dto_data_source_type import DatasourceDtoDataSourceType
from ..models.datasource_dto_status import DatasourceDtoStatus
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.datasource_config import DatasourceConfig
    from ..models.datasource_dto_added_by import DatasourceDtoAddedBy
    from ..models.datasource_owner_info import DatasourceOwnerInfo


T = TypeVar("T", bound="DatasourceDto")


@_attrs_define
class DatasourceDto:
    """
    Attributes:
        added_by (DatasourceDtoAddedBy):
        auth_type (DatasourceDtoAuthType):
        config (DatasourceConfig):
        data_source_type (DatasourceDtoDataSourceType):
        id (str):
        last_updated (int):
        last_updated_by (str):
        name (str):
        owner_id (str):
        owner_info (DatasourceOwnerInfo):
        project_ids (List[str]):
        status (DatasourceDtoStatus):
        description (Union[None, Unset, str]):
    """

    added_by: "DatasourceDtoAddedBy"
    auth_type: DatasourceDtoAuthType
    config: "DatasourceConfig"
    data_source_type: DatasourceDtoDataSourceType
    id: str
    last_updated: int
    last_updated_by: str
    name: str
    owner_id: str
    owner_info: "DatasourceOwnerInfo"
    project_ids: List[str]
    status: DatasourceDtoStatus
    description: Union[None, Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        added_by = self.added_by.to_dict()

        auth_type = self.auth_type.value

        config = self.config.to_dict()

        data_source_type = self.data_source_type.value

        id = self.id

        last_updated = self.last_updated

        last_updated_by = self.last_updated_by

        name = self.name

        owner_id = self.owner_id

        owner_info = self.owner_info.to_dict()

        project_ids = self.project_ids

        status = self.status.value

        description: Union[None, Unset, str]
        if isinstance(self.description, Unset):
            description = UNSET
        else:
            description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "addedBy": added_by,
                "authType": auth_type,
                "config": config,
                "dataSourceType": data_source_type,
                "id": id,
                "lastUpdated": last_updated,
                "lastUpdatedBy": last_updated_by,
                "name": name,
                "ownerId": owner_id,
                "ownerInfo": owner_info,
                "projectIds": project_ids,
                "status": status,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.datasource_config import DatasourceConfig
        from ..models.datasource_dto_added_by import DatasourceDtoAddedBy
        from ..models.datasource_owner_info import DatasourceOwnerInfo

        d = src_dict.copy()
        added_by = DatasourceDtoAddedBy.from_dict(d.pop("addedBy"))

        auth_type = DatasourceDtoAuthType(d.pop("authType"))

        config = DatasourceConfig.from_dict(d.pop("config"))

        data_source_type = DatasourceDtoDataSourceType(d.pop("dataSourceType"))

        id = d.pop("id")

        last_updated = d.pop("lastUpdated")

        last_updated_by = d.pop("lastUpdatedBy")

        name = d.pop("name")

        owner_id = d.pop("ownerId")

        owner_info = DatasourceOwnerInfo.from_dict(d.pop("ownerInfo"))

        project_ids = cast(List[str], d.pop("projectIds"))

        status = DatasourceDtoStatus(d.pop("status"))

        def _parse_description(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        description = _parse_description(d.pop("description", UNSET))

        datasource_dto = cls(
            added_by=added_by,
            auth_type=auth_type,
            config=config,
            data_source_type=data_source_type,
            id=id,
            last_updated=last_updated,
            last_updated_by=last_updated_by,
            name=name,
            owner_id=owner_id,
            owner_info=owner_info,
            project_ids=project_ids,
            status=status,
            description=description,
        )

        datasource_dto.additional_properties = d
        return datasource_dto

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.feature_store_sync_result import FeatureStoreSyncResult

T = TypeVar("T", bound="UnlockFeatureStoreRequest")


@_attrs_define
class UnlockFeatureStoreRequest:
    """
    Attributes:
        feature_store_id (str):
        sync_result (FeatureStoreSyncResult):
    """

    feature_store_id: str
    sync_result: FeatureStoreSyncResult
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        feature_store_id = self.feature_store_id

        sync_result = self.sync_result.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "featureStoreId": feature_store_id,
                "syncResult": sync_result,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        feature_store_id = d.pop("featureStoreId")

        sync_result = FeatureStoreSyncResult(d.pop("syncResult"))

        unlock_feature_store_request = cls(
            feature_store_id=feature_store_id,
            sync_result=sync_result,
        )

        unlock_feature_store_request.additional_properties = d
        return unlock_feature_store_request

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class LogMetricT(str, Enum):
    ADLSCONFIG = "ADLSConfig"
    AZUREBLOBSTORAGECONFIG = "AzureBlobStorageConfig"
    GCSCONFIG = "GCSConfig"
    GENERICS3CONFIG = "GenericS3Config"
    S3CONFIG = "S3Config"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, Optional, Union

from http import HTTPStatus

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.datasource_dto import DatasourceDto
from ...models.error_response import ErrorResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    name: str,
    *,
    run_id: Union[Unset, str] = UNSET,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}

    params["runId"] = run_id

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "get",
        "url": "/datasource/name/{name}".format(
            name=name,
        ),
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[DatasourceDto, ErrorResponse]]:
    if response.status_code == 200:
        response_200 = DatasourceDto.from_dict(response.json())

        return response_200
    if response.status_code == 400:
        response_400 = ErrorResponse.from_dict(response.json())

        return response_400
    if response.status_code == 401:
        response_401 = ErrorResponse.from_dict(response.json())

        return response_401
    if response.status_code == 403:
        response_403 = ErrorResponse.from_dict(response.json())

        return response_403
    if response.status_code == 404:
        response_404 = ErrorResponse.from_dict(response.json())

        return response_404
    if response.status_code == 500:
        response_500 = ErrorResponse.from_dict(response.json())

        return response_500
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[DatasourceDto, ErrorResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    name: str,
    *,
    client: Union[AuthenticatedClient, Client],
    run_id: Union[Unset, str] = UNSET,
) -> Response[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[DatasourceDto, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        name=name,
        run_id=run_id,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    name: str,
    *,
    client: Union[AuthenticatedClient, Client],
    run_id: Union[Unset, str] = UNSET,
) -> Optional[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[DatasourceDto, ErrorResponse]
    """

    return sync_detailed(
        name=name,
        client=client,
        run_id=run_id,
    ).parsed


async def asyncio_detailed(
    name: str,
    *,
    client: Union[AuthenticatedClient, Client],
    run_id: Union[Unset, str] = UNSET,
) -> Response[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[DatasourceDto, ErrorResponse]]
    """

    kwargs = _get_kwargs(
        name=name,
        run_id=run_id,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    name: str,
    *,
    client: Union[AuthenticatedClient, Client],
    run_id: Union[Unset, str] = UNSET,
) -> Optional[Union[DatasourceDto, ErrorResponse]]:
    """Get datasource by name

    Args:
        name (str):
        run_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[DatasourceDto, ErrorResponse]
    """

    return (
        await asyncio_detailed(
            name=name,
            client=client,
            run_id=run_id,
        )
    ).parsed
